# -*- coding: utf-8 -*-

import os
import shutil
import sys
import tempfile
import time
import unittest

from dirq import queue
from dirq.queue import Queue, QueueError

__all__ = ['TestQueue', 'TestQueueModuleFunctions']


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq')
        self.qdir = self.tempdir + '/dirq'

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueue(TestDirQueue):

    def test1init(self):
        """ Queue.__init__() """
        path = self.tempdir + '/aaa/bbb/ccc'
        umask = None
        maxelts = 10
        good_schemas = [
            {'a': 'binary'},
            {'a': 'binary*'},
            {'a': 'string'},
            {'a': 'string*'},
            {'a': 'table'},
            {'a': 'binary?', 'b': 'binary'},
            {'a': 'binary?', 'b': 'binary*'},
            {'a': 'string', 'b': 'binary?*'}, ]
        for schema in good_schemas:
            try:
                Queue(path, umask=umask, maxelts=maxelts, schema=schema)
            except Exception:
                error = sys.exc_info()[1]
                self.fail("Shouldn't have failed. Exception: %s" % error)
        bad_schemas = [['foo'],
                       {'foo': 1},
                       {'a': 'binary?'}]
        for schema in bad_schemas:
            self.assertRaises(
                QueueError, Queue, path,
                umask=umask, maxelts=maxelts, schema=schema)
        bad_schemas = [{'a': 'strings'}, {'a': 'table??'}]
        for schema in bad_schemas:
            self.assertRaises(
                QueueError, Queue, path,
                umask=umask, maxelts=maxelts, schema=schema)

    def test2copy(self):
        """ Queue.copy(). """
        q = Queue(self.qdir, schema={'a': 'string'})
        q1 = q.copy()
        q.foo = 1
        try:
            q1.foo
        except AttributeError:
            pass
        else:
            self.fail('Not a copy, but reference.')

    def test3_is_locked(self):
        """ Queue._is_locked_*() """
        q = Queue(self.qdir, schema={'a': 'string'})

        assert q._is_locked_nlink('') is False
        assert q._is_locked_nlink('not_there') is False
        os.mkdir(self.qdir + '/a')
        assert q._is_locked_nlink('a') is False
        os.mkdir(self.qdir + '/a/%s' % queue.LOCKED_DIRECTORY)
        assert q._is_locked_nlink('a') is True
        time.sleep(1)
        assert q._is_locked_nlink('a', time.time()) is True

        assert q._is_locked_nonlink('') is False
        assert q._is_locked_nonlink('not_there') is False
        os.mkdir(self.qdir + '/b')
        assert q._is_locked_nonlink('b') is False
        os.mkdir(self.qdir + '/b/%s' % queue.LOCKED_DIRECTORY)
        assert q._is_locked_nonlink('b') is True
        time.sleep(1)
        assert q._is_locked_nonlink('b', time.time()) is True

    def test4_insertion_directory(self):
        """ Queue._insertion_directory() """
        q = queue.Queue(self.qdir, schema={'a': 'string'})
        q.maxelts = 1
        name0 = '%08x' % 0
        assert q._insertion_directory() == name0
        assert os.path.exists(self.qdir + '/' + name0)
        os.mkdir('%s/%s/%s' % (self.qdir, name0, queue._name(q.rndhex)))
        name1 = '%08x' % 1
        assert q._insertion_directory() == name1
        assert os.path.exists(self.qdir + '/' + name1)

    def test5add(self):
        """ Queue.add() """
        q = queue.Queue(self.qdir, schema={'a': 'string'})
        q.add({'a': 'a\n'})

        assert os.listdir(self.qdir + '/' + queue.TEMPORARY_DIRECTORY) == []
        data_file = '%s/%s/%s/a' % (self.qdir, '%08x' % 0,
                                    os.listdir('%s/%08x' % (self.qdir, 0))[-1])
        assert os.path.exists(data_file)
        assert open(data_file).read() == 'a\n'

    def test6touch(self):
        """ Queue.touch() """
        q = queue.Queue(self.qdir, schema={'a': 'string'})
        q.add({'a': 'a\n'})
        e = q.first()
        element_dir = q.path + '/' + e
        old_time = time.time() - 10
        os.utime(element_dir, (old_time, old_time))
        mtime = os.stat(element_dir).st_mtime
        q.touch(e)
        assert os.stat(element_dir).st_mtime > mtime


class TestQueueModuleFunctions(TestDirQueue):

    def test2_check_element(self):
        """ queue._check_element() """
        queue._check_element('0' * 8 + '/' + '0' * 14)
        queue._check_element('f' * 8 + '/' + 'f' * 14)
        for e in ['f' * 7 + '/' + 'f' * 14,
                  'f' * 9 + '/' + 'f' * 14,
                  'f' * 8 + '/' + 'f' * 13,
                  'f' * 8 + '/' + 'f' * 15,
                  'f' * 8 + '/' + 'g' * 14,
                  'g' * 8 + '/' + 'f' * 14,
                  ]:
            self.assertRaises(queue.QueueError, queue._check_element, (e))

    def test3_hash2string(self):
        """ queue._hash2string() """
        assert queue._hash2string({'a1': 'a2'}) == 'a1\ta2\n'
        assert queue._hash2string({'a1\\': 'a2'}) == 'a1\\\\\ta2\n'
        assert queue._hash2string({'a1	a2': 'a3	a4'}) == 'a1\\ta2\ta3\\ta4\n'
        assert queue._hash2string({'a1	\na2': 'a3	\na4'}) == \
            'a1\\t\\na2\ta3\\t\\na4\n'

    def test3_string2hash(self):
        """ queue._string2hash() """
        assert queue._string2hash('a1\ta2\nb1\tb2') == {'a1': 'a2', 'b1': 'b2'}
        assert queue._string2hash('a1\x5c\ta2\nb1\tb2\\') == \
            {'a1\x5c': 'a2', 'b1': 'b2\\'}
        for value in ['a', ]:
            self.assertRaises(queue.QueueError, queue._string2hash, (value))

    def test3_hash2string2hash(self):
        """ queue._hash2string()+queue._string2hash() """
        example = {"hi\\t\th\nere": "h\\ello\twor\nld"}
        converted = queue._hash2string(example)
        assert converted == "hi\\\\t\\th\\nere\th\\\\ello\\twor\\nld\n"
        back = queue._string2hash(converted)
        assert back == example


def main():
    """ Main. """
    testcases = [TestQueue,
                 TestQueueModuleFunctions]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



"""Directory based queue.

A port of Perl module Directory::Queue
http://search.cpan.org/dist/Directory-Queue/

The documentation from Directory::Queue module was adapted for Python.

The goal of this module is to offer a queue system using the underlying
filesystem for storage, security and to prevent race conditions via atomic
operations. It focuses on simplicity, robustness and scalability.

This module allows multiple concurrent readers and writers to interact with
the same queue.

For usage and implementation details see :py:mod:`dirq.queue` module.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

AUTHOR = "Konstantin Skaburskas <konstantin.skaburskas@gmail.com>"
VERSION = "1.8"
DATE = "3 Nov 2021"
__author__ = AUTHOR
__version__ = VERSION
__date__ = DATE



"""Base class and common code for :py:mod:`dirq` package.

It is used internally by :py:mod:`dirq` modules and should not
be used elsewhere.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import dirq
import codecs
import errno
import inspect
import os
import random
import re
import sys
import time

from dirq.utils import VALID_STR_TYPES

__author__ = dirq.AUTHOR
__version__ = dirq.VERSION
__date__ = dirq.DATE

__DIRECTORY_REGEXP = '[0-9a-f]{8}'
__ELEMENT_REGEXP = '[0-9a-f]{14}'
_DIRECTORY_REGEXP = re.compile('^(%s)$' % __DIRECTORY_REGEXP)
_ELEMENT_REGEXP = re.compile('^(%s)$' % __ELEMENT_REGEXP)
_DIRELT_REGEXP = re.compile('^%s/%s$' % (__DIRECTORY_REGEXP, __ELEMENT_REGEXP))

WARN = False


def _warn(text):
    """ Print a warning. """
    if WARN:
        sys.stdout.write('%s, at %s line %s\n' %
                         (text, __name__,
                          inspect.currentframe().f_back.f_lineno))
        sys.stdout.flush()


def _name(rndhex):
    """
    Return the name of a new element to (try to) use with:
    * 8 hexadecimal digits for the number of seconds since the Epoch
    * 5 hexadecimal digits for the microseconds part
    * 1 hexadecimal digit from the pid to further reduce name collisions

    Properties:
    * fixed size (14 hexadecimal digits)
    * likely to be unique (with high-probability)
    * can be lexically sorted
    * ever increasing (for a given process)
    * reasonably compact
    * matching _ELEMENT_REGEXP
    """
    now = time.time()
    secs = int(now)
    msecs = int((now - secs) * 1000000)
    return "%08x%05x%01x" % (secs, msecs, rndhex)


def _directory_contents(path, missingok=True):
    """Get the contents of a directory as a list of names, without . and ..

    Raise:
        OSError - can't list directory

    Note:
    * if the optional second argument is true, it is not an error if the
      directory does not exist (anymore)
    """
    try:
        return os.listdir(path)
    except Exception:
        error = sys.exc_info()[1]
        if not missingok and not error.errcode == errno.ENOENT:
            raise OSError("cannot listdir(%s): %s" % (path, error))
            # RACE: this path does not exist (anymore)
        return []


def _wrapped_makedirs(path):
    """Wrapped os.makedirs() used by _special_mkdir()"""
    try:
        os.makedirs(path)
    except OSError:
        error = sys.exc_info()[1]
        if error.errno == errno.EEXIST and os.path.isdir(path):
            return (False, None)
        elif error.errno == errno.EISDIR:
            return (False, None)
        return (False, "cannot mkdir(%s): %s" % (path, error))
    else:
        return (True, None)


def _special_mkdir(path, umask=None):
    """
    Recursively create directories specified in path:
    * return true on success
    * return false if the directory already exists
    * die in case of any other error

    Raise:
        OSError - can't make directory
    """
    if umask is None:
        result, error = _wrapped_makedirs(path)
    else:
        oldumask = os.umask(umask)
        result, error = _wrapped_makedirs(path)
        os.umask(oldumask)
    if error is None:
        return result
    raise OSError(error)


def _special_rmdir(path):
    """
    Delete a directory:
    * return true on success
    * return false if the path does not exist (anymore)
    * die in case of any other error

    Raise:
        OSError - can't delete given directory
    """
    try:
        os.rmdir(path)
    except Exception:
        error = sys.exc_info()[1]
        if not error.errno == errno.ENOENT:
            raise OSError("cannot rmdir(%s): %s" % (path, error))
            # RACE: this path does not exist (anymore)
        return False
    else:
        return True


def _file_read(path, utf8):
    """Read from a file.

    Raise:
        OSError - problems opening/closing file
        IOError - file read error
    """
    try:
        if utf8:
            fileh = codecs.open(path, 'r', "utf8")
        else:
            fileh = open(path, 'rb')
    except Exception:
        error = sys.exc_info()[1]
        raise OSError("cannot open %s: %s" % (path, error))
    try:
        data = fileh.read()
    except Exception:
        error = sys.exc_info()[1]
        raise IOError("cannot read %s: %s" % (path, error))
    try:
        fileh.close()
    except Exception:
        error = sys.exc_info()[1]
        raise OSError("cannot close %s: %s" % (path, error))
    return data


def _file_create(path, umask=None, utf8=False):
    """Open a file defined by 'path' and return file handler.

    Raises:
        OSError - if file exists
    """
    if umask is not None:
        oldumask = os.umask(umask)
    if utf8:
        if os.path.exists(path):
            ex = OSError("[Errno %i] File exists: %s" % (errno.EEXIST, path))
            ex.errno = errno.EEXIST
            raise ex
        fileh = codecs.open(path, 'w', 'utf8')
    else:
        fileh = os.fdopen(
            os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 438), 'wb')
    if umask is not None:
        os.umask(oldumask)

    return fileh


def _file_write(path, utf8, umask, data):
    """Write to a file.

    Raise:
        OSError - problems opening/closing file
        IOError - file write error
    """
    fileh = _file_create(path, umask=umask, utf8=utf8)
    try:
        fileh.write(data)
    except Exception:
        error = sys.exc_info()[1]
        raise IOError("cannot write to %s: %s" % (path, error))
    try:
        fileh.close()
    except Exception:
        error = sys.exc_info()[1]
        raise OSError("cannot close %s: %s" % (path, error))


class QueueBase(object):
    """QueueBase
    """
    def __init__(self, path, umask=None, rndhex=None):
        """
        Arguments:
            path
                the queue toplevel directory
            umask
                the umask to use when creating files and directories
                (default: use the running process' umask)
            rndhex
                the hexadecimal digit to use in names
                (default: randomly chosen)

        Raise:
            TypeError  - wrong input data types provided
            OSError    - can't create directory structure
        """
        self.dirs = []
        self.elts = []
        self._next_exception = False

        if type(path) not in VALID_STR_TYPES:
            raise TypeError("'path' should be str or unicode")
        self.path = path
        if umask is not None and not isinstance(umask, int):
            raise TypeError("'umask' should be integer")
        self.umask = umask
        if rndhex is not None and not isinstance(rndhex, int):
            raise TypeError("'rndhex' should be integer")
        if rndhex is None:
            self.rndhex = random.randint(0, 15)
        else:
            self.rndhex = rndhex % 16

        # create top level directory
        _special_mkdir(path, self.umask)

        # store the queue unique identifier
        if sys.platform in ['win32']:
            self.id = self.path
        else:
            stat = os.stat(self.path)
            self.id = '%s:%s' % (stat.st_dev, stat.st_ino)

    def __iter__(self):
        """Return iterator over element names.
        """
        self._reset()
        self._next_exception = True
        return self

    def names(self):
        """Return iterator over element names.
        """
        return self.__iter__()

    def copy(self):
        """Copy/clone the object. Return copy of the object.

        Note:

        * the main purpose is to copy/clone the iterator cached state
        * the other structured attributes (including schema) are not cloned
        """
        import copy
        new = copy.deepcopy(self)
        new.dirs = []
        new.elts = []
        return new

    def _reset(self):
        """Regenerate list of intermediate directories. Drop cached
        elements list.

        Raise:
            OSError - can't list directories
        """
        self.dirs = []
        for name in _directory_contents(self.path):
            if _DIRECTORY_REGEXP.match(name):
                self.dirs.append(name)
        self.dirs.sort()
        self.elts = []

    def first(self):
        """Return the first element in the queue and cache information about
        the next ones.

        Raise:
            OSError - can't list directories
        """
        self._reset()
        return self.next()

    def _build_elements(self):
        """ This should be implemented by sub classes. """
        raise NotImplementedError('Implement in sub-class.')

    def __next__(self):
        """Return name of the next element in the queue, only using cached
        information. When queue is empty, depending on the iterator
        protocol - return empty string or raise StopIteration.

        Return:
            name of the next element in the queue

        Raise:
            StopIteration - when used as Python iterator via
                            __iter__() method
        """
        if self.elts:
            return self.elts.pop(0)
        if self._build_elements():
            return self.elts.pop(0)
        if self._next_exception:
            self._next_exception = False
            raise StopIteration
        else:
            return ''
    next = __next__

    def touch(self, ename):
        """Touch an element directory to indicate that it is still being used.

        Note:
            this is only really useful for locked elements but we allow it
            for all.

        Raises:
           EnvironmentError - on any IOError, OSError in utime()

        NOTE: this may not work on OSes with directories implemented not as
        files (eg. Windows). See doc for os.utime().
        """
        path = '%s/%s' % (self.path, ename)
        try:
            os.utime(path, None)
        except (IOError, OSError):
            error = sys.exc_info()[1]
            raise EnvironmentError("cannot utime(%s, None): %s" %
                                   (path, error))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test program for testing the dirq.QueueSet module.
"""

import os
import re
import shutil
import sys
import time

from optparse import OptionParser

sys.path.insert(1, re.sub(r'/\w*$', '', os.getcwd()))

import dirq  # noqa E402
from dirq import queue  # noqa E402
from dirq.QueueSimple import QueueSimple  # noqa E402

opts = None
TEST = ''
ProgramName = sys.argv[0]


def init():
    """ Initialize. """
    global opts, TEST
    parser = OptionParser(usage="%prog [OPTIONS] [--] TEST",
                          version=("%%prog %s" % dirq.VERSION))
    parser.add_option('-d', '--debug', dest='debug', action="store_true",
                      default=False, help="show debugging information")
    parser.add_option('-p', '--path', dest='path', type='string', default='',
                      help="set the queue path")
    parser.add_option('-c', '--count', dest='count', type='int', default=0,
                      help="set the elements count")
    parser.add_option("-s", "--size", dest="size", type='int', default=0,
                      help="set the body size for added elements")
    parser.add_option("-r", "--random", dest="random", action="store_true",
                      default=False, help="randomize the body size")
    parser.add_option("--granularity", dest="granularity", type="int",
                      default=None, help="time granularity for intermediate "
                      "directories (QueueSimple)")
    parser.add_option("--header", dest="header", action="store_true",
                      default=False, help="set header for added elements")
    parser.add_option("--maxelts", dest="maxelts", type='int', default=0,
                      help="set the maximum number of elements per directory")
    parser.add_option("--type", dest="type", type="string", default="simple",
                      help="set the type of dirq (simple|normal)")
    opts, args = parser.parse_args()
    if not opts.path:
        _die("%s: mandatory option not set: -p/--path", ProgramName)
    if len(args) != 0:
        TEST = args[0]
    else:
        parser.print_help()
        sys.exit()


def debug(fmt, *arguments):
    """Report a debugging message.
    """
    if not opts.debug:
        return
    message = fmt % arguments
    message = re.sub(r'\s+$', '.', message)
    sys.stderr.write("# %s %s[%d]: %s\n" %
                     (time.strftime("%Y/%m/%d-%H:%M:%S",
                                    time.localtime(time.time())),
                      os.path.basename(sys.argv[0]),
                      os.getpid(), message))


def _die(fmt, *arguments):
    """Report a fatal error."""
    sys.stderr.write(fmt % arguments + "\n")
    sys.stderr.flush()
    sys.exit(1)


def new_dirq(path, _schema):
    """Create a new Directory::Queue object, optionally with schema.
    """
    kwargs = {}
    if opts.type == "simple":
        if opts.granularity is not None:
            kwargs['granularity'] = opts.granularity
        return QueueSimple(path, **kwargs)
    else:
        if _schema:
            schema = {'body': 'string',
                      'header': 'table?'}
            kwargs['schema'] = schema
        if opts.maxelts:
            kwargs['maxelts'] = opts.maxelts
        return queue.Queue(path, **kwargs)


def new_dirqs():
    """Create a new Directory::Queue object, optionally with schema.
    """
    time1 = time.time()
    qs = queue.QueueSet()
    for path in opts.path.split(','):
        qs.add(new_dirq(path, 0))
    debug("created queue set in %.4f seconds", time.time() - time1)
    return qs


def test_count():
    """Count the elements in the queue.
    """
    qs = new_dirqs()
    time1 = time.time()
    count = qs.count()
    time2 = time.time()
    debug("queue set has %d elements", count)
    debug("done in %.4f seconds", time2 - time1)


def test_add():
    """Add elements to the queue.
    """


def test_complex():
    """Add elements to the queue.
    """
    wd = opts.path
    os.mkdir(wd)
    qn = 6
    paths = []
    for i in range(qn):
        paths.append(wd + '/q%i' % i)
    count = opts.count or 1000
    debug("creating %i initial queues. adding %i elements into each." %
          (qn, count))
    queues = []
    t1 = time.time()
    while qn:
        dq = new_dirq(paths[qn - 1], 1)
        debug("adding %d elements to the queue...", count)
        element = {}
        done = 0
        time1 = time.time()
        while not count or done < count:
            done += 1
            element['body'] = 'Element %i \u263A\n' % done
            dq.add(element)
        time2 = time.time()
        debug("done in %.4f seconds", time2 - time1)
        queues.append(dq)
        qn -= 1
    debug("total done in %.4f seconds", time.time() - t1)

    time1 = time.time()
    i = 3
    qs = queue.QueueSet(queues[0:i])
    debug("created queue set in %.4f seconds", time.time() - time1)
    debug("elements in %i queues: %i" % (i, qs.count()))

    debug("adding remaining queues to the set.")
    t1 = time.time()
    qs.add(queues[i:])
    debug("done in %.4f sec." % (time.time() - t1))
    debug("total element with added queues: %i" % qs.count())

    debug("removing %i first queues." % i)
    t1 = time.time()
    for dq in queues[0:i]:
        qs.remove(dq)
    debug("done in %.4f sec." % (time.time() - t1))

    debug("number of elements left: %i" % qs.count())

    debug("deleting queues from disk...")
    for path in paths:
        shutil.rmtree(path, ignore_errors=True)
    debug("done.")


def test_iterate():
    """Iterate through the set of queues (only lock+unlock).
    """
    debug("iterating all elements in the set of queues (one pass)...")
    qs = new_dirqs()
    done = 0
    time1 = time.time()
    dq, name = qs.first()
    while dq:
        if not dq.lock(name):
            dq, name = qs.next()
            continue
        dq.unlock(name)
        done += 1
        dq, name = qs.next()
    time2 = time.time()
    debug("done in %.4f seconds (%d elements)", time2 - time1, done)


if __name__ == "__main__":
    init()
    if TEST == "count":
        test_count()
    elif TEST == "add":
        test_add()
    elif TEST == "complex":
        test_complex()
    elif TEST == "iterate":
        test_iterate()
    else:
        _die("%s: unsupported test: %s", ProgramName, TEST)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Producer, browser and consumer. """

from dirq.queue import Queue, QueueError, QueueLockError
import os
import sys
import tempfile

# total number of elements
COUNT = 9
# queue head directory
path = tempfile.mkdtemp()
# max elements per elements directory
maxelts = 3
# element's schema
schema = {'body': 'string',
          'header': 'table?'}

# ========
# PRODUCER
print("*** PRODUCER")
dirq_p = Queue(path, maxelts=maxelts, schema=schema)

print("adding %d elements to the queue at %s" % (COUNT, path))
done = 1
while done <= COUNT:
    element = {}
    try:
        element['body'] = ('Élément %i \u263A\n' % done).decode("utf-8")
    except AttributeError:
        element['body'] = 'Élément %i \u263A\n' % done
    if done % 2:  # header only for odd sequential elements
        element['header'] = dict(os.environ)
    name = dirq_p.enqueue(element)
    # name = dirq_p.add(element) # same
    print("added %.2i: %s" % (done, name))
    done += 1

total_stored = dirq_p.count()
print("total elements in the queue: %d" % total_stored)
assert total_stored == COUNT
del dirq_p
print('=' * 25)

# =======
# BROWSER
print("*** BROWSER: Python iterator protocol.")
dirq_b = Queue(path, schema=schema)
done = 0
miss = 0
for i, name in enumerate(dirq_b.names()):
    done += 1  # enumerate(o, start=1) in Python 2.6
    print("element: %s %s" % (path, name))
    try:
        if i in [1, 4, 5]:  # artificially lock some elements
            os.mkdir('%s/%s/locked' % (path, name))
        e = dirq_b.get_element(name)  # lock(name), get(name), unlock(name)
    except QueueLockError:
        error = sys.exc_info()[1]
        print(error)
        miss += 1
        continue
    except QueueError:
        error = sys.exc_info()[1]
        print(error)
        miss += 1
        continue
    except Exception:
        error = sys.exc_info()[1]
        print(error)
        break
    print("element: %i" % i, e.keys())
print("found %i elements; got %i" % (done, done - miss))
del dirq_b

print("*** BROWSER: lower level iterator protocol.")
dirq_b = Queue(path, schema=schema)
done = 0
miss = 0
name = dirq_b.first()
while name:
    print("element: %s %s" % (path, name))
    if not dirq_b.lock(name):
        name = dirq_b.next()
        miss += 1
        continue
    element = dirq_b.get(name)
    print("element: %i" % done, element.keys())
    dirq_b.unlock(name)
    done += 1
    name = dirq_b.next()
print("found %i elements; got %i" % (done + miss, done))
assert done == (COUNT - miss)
del dirq_b
print('=' * 25)

# ========
# CONSUMER
print("*** CONSUMER: Python iterator protocol.")
dirq_c = Queue(path, schema=schema)
miss = 0
for i, name in enumerate(dirq_c):
    i += 1  # enumerate(o, start=1) in Python 2.6
    print("element %i: %s %s" % (i, path, name))
    try:
        e = dirq_c.dequeue(name)  # lock(name), get(name), remove(name)
    except QueueError:
        error = sys.exc_info()[1]
        print(error)
        miss += 1
        continue
    except Exception:
        error = sys.exc_info()[1]
        print(error)
        break
    print(e.keys())
print("consumed %i elements out of %i seen" % (i - miss, i))
total_left = dirq_c.count()
print("elements left in the queue: %d" % total_left)
assert total_left == miss
del dirq_c

print("*** CONSUMER: lower level iterator protocol.")
path = tempfile.mkdtemp()
print("create new queue and add elements")
print("adding %d elements to the queue at %s" % (COUNT, path))
dirq_p = Queue(path, maxelts=maxelts, schema=schema)
done = 1
while done <= COUNT:
    element = {}
    try:
        element['body'] = ('Élément %i \u263A\n' % done).decode("utf-8")
    except AttributeError:
        element['body'] = 'Élément %i \u263A\n' % done
    if done % 2:  # header only for odd sequential elements
        element['header'] = dict(os.environ)
    name = dirq_p.enqueue(element)
    print("added %.2i: %s" % (done, name))
    done += 1

print("start consuming...")
dirq_c = Queue(path, schema=schema)
done = 0
name = dirq_c.first()
while name:
    print("element: %s %s" % (path, name))
    if not dirq_c.lock(name):
        print("couldn't lock: %s" % name)
        name = dirq_c.next()
        continue
    element = dirq_c.get(name)
    print(element.keys())
    dirq_c.remove(name)
    done += 1
    name = dirq_c.next()
print("consumed %i elements" % done)
total_left = dirq_c.count()
print("elements left in the queue: %d" % total_left)
assert total_left == 0



"""Directory based queue.

A port of Perl module Directory::Queue
http://search.cpan.org/dist/Directory-Queue/
The documentation from Directory::Queue module was adapted for Python.

The goal of this module is to offer a simple queue system using the
underlying filesystem for storage, security and to prevent race
conditions via atomic operations. It focuses on simplicity, robustness
and scalability.

This module allows multiple concurrent readers and writers to interact
with the same queue.

Provides
========

Classes:

* :py:class:`dirq.queue.Queue`       directory based queue
* :py:class:`dirq.QueueSimple.QueueSimple` simple directory based queue
* :py:class:`dirq.queue.QueueSet`    set of directory based queues
* :py:class:`dirq.Exceptions.QueueError`  exception

Documentation
=============

===========
Queue class
===========

:py:class:`dirq.queue.Queue` - directory based queue.

Usage::

    from dirq.queue import Queue

    # simple schema:
    #  - there must be a "body" which is a string
    #  - there can be a "header" which is a table/dictionary

    schema = {"body": "string", "header": "table?"}
    queuedir = "/tmp/test"

    # sample producer

    dirq = Queue(queuedir, schema=schema)
    import os
    for count in range(1,101):
        name = dirq.add({"body"  : "element %i"%count,
                         "header": dict(os.environ)})
        print("# added element %i as %s" % (count, name))

    # sample consumer

    dirq = Queue(queuedir, schema=schema)
    name = dirq.first()
    while name:
        if not dirq.lock(name):
            name = dirq.next()
            continue
        print("# reading element %s" % name)
        data = dirq.get(name)
        # one can use data['body'] and data['header'] here...
        # one could use dirq.unlock(name) to only browse the queue...
        dirq.remove(name)
        name = dirq.next()

Terminology
-----------

    An element is something that contains one or more pieces of data. A
    simple string may be an element but more complex schemas can also be
    used, see the *Schema* section for more information.

    A queue is a "best effort FIFO" collection of elements.

    It is very hard to guarantee pure FIFO behavior with multiple writers
    using the same queue. Consider for instance:

    . Writer1: calls the add() method
    . Writer2: calls the add() method
    . Writer2: the add() method returns
    . Writer1: the add() method returns

    Who should be first in the queue, Writer1 or Writer2?

    For simplicity, this implementation provides only "best effort FIFO",
    i.e. there is a very high probability that elements are processed in
    FIFO order but this is not guaranteed. This is achieved by using a
    high-resolution time function and having elements sorted by the time
    the element's final directory gets created.

Locking
-------

    Adding an element is not a problem because the add() method is atomic.

    In order to support multiple processes interacting with the same queue,
    advisory locking is used. Processes should first lock an element before
    working with it. In fact, the get() and remove() methods raise an
    exception if they are called on unlocked elements.

    If the process that created the lock dies without unlocking the ele-
    ment, we end up with a staled lock. The purge() method can be used to
    remove these staled locks.

    An element can basically be in only one of two states: locked or
    unlocked.

    A newly created element is unlocked as a writer usually does not need
    to do anything more with the element once dropped in the queue.

    Iterators return all the elements, regardless of their states.

    There is no method to get an element state as this information is usu-
    ally useless since it may change at any time. Instead, programs should
    directly try to lock elements to make sure they are indeed locked.

Constructor
-----------

    For the signature of the Queue constructor see documentation to the
    respective __init__() method.

Schema
------
    The schema defines how user supplied data is stored in the queue. It is
    only required by the add() and get() methods.

    The schema must be a dictionary containing key/value pairs.

    The key must contain only alphanumerical characters. It identifies the
    piece of data and will be used as file name when storing the data
    inside the element directory.

    The value represents the type of the given piece of data. It can be:

    binary
        the data is a sequence of binary bytes, it will be stored directly
        in a plain file with no further encoding

    string
        the data is a text string (i.e. a sequence of characters), it will
        be UTF-8 encoded

    table
        the data is a reference to a hash of text strings, it will be seri-
        alized and UTF-8 encoded before being stored in a file

    By default, all pieces of data are mandatory. If you append a question
    mark to the type, this piece of data will be marked as optional. See
    the comments in the *Usage* section for more information.

    To comply with Directory::Queue implementation it is allowed to
    append '*' (asterisk) to data type specification, which in
    Directory::Queue means switching to working with element references in
    add() and get() operations. This is irrelevant for the Python
    implementation.

Directory Structure
-------------------

    All the directories holding the elements and all the files holding the
    data pieces are located under the queue toplevel directory. This direc-
    tory can contain:

    temporary
        the directory holding temporary elements, i.e. the elements being
        added

    obsolete
        the directory holding obsolete elements, i.e. the elements being
        removed

    NNNNNNNN
        an intermediate directory holding elements; NNNNNNNN is an 8-digits
        long hexadecimal number

    In any of the above directories, an element is stored as a single
    directory with a 14-digits long hexadecimal name SSSSSSSSMMMMMR where:

    SSSSSSSS
        represents the number of seconds since the Epoch

    MMMMM
        represents the microsecond part of the time since the Epoch

    R   is a random digit used to reduce name collisions

    Finally, inside an element directory, the different pieces of data are
    stored into different files, named according to the schema. A locked
    element contains in addition a directory named "locked".

Security
--------

    There are no specific security mechanisms in this module.

    The elements are stored as plain files and directories. The filesystem
    security features (owner, group, permissions, ACLs...) should be used
    to adequately protect the data.

    By default, the process' umask is respected. See the class constructor
    documentation if you want an other behavior.

    If multiple readers and writers with different uids are expected, the
    easiest solution is to have all the files and directories inside the
    toplevel directory world-writable (i.e. umask=0). Then, the permissions
    of the toplevel directory itself (e.g. group-writable) are enough to
    control who can access the queue.


==============
QueueSet class
==============

:py:class:`dirq.queue.QueueSet` - interface to a set of Queue objects

Usage::

    from dirq.queue import Queue, QueueSet

    dq1 = Queue("/tmp/q1")
    dq2 = Queue("/tmp/q2")
    dqset = QueueSet(dq1, dq2)
    # dqs = [dq1, dq2]
    # dqset = QueueSet(dqs)

    (dq, elt) = dqset.first()
    while dq:
        # you can now process the element elt of queue dq...
        (dq, elt) = dqset.next()

Description
-----------

    This class can be used to put different queues into a set and browse
    them as one queue. The elements from all queues are merged together
    and sorted independently from the queue they belong to.

Constructor
-----------

    For the signature of the QueueSet constructor see documentation to the
    respective :py:meth:`dirq.queue.QueueSet.__init__` method.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import dirq
import errno
import os
import re
import sys
import time

from dirq.QueueBase import QueueBase, _DIRELT_REGEXP, _DIRECTORY_REGEXP, \
    _ELEMENT_REGEXP
from dirq.QueueBase import (
    _name,
    _special_mkdir,
    _special_rmdir,
    _file_read,
    _file_write,
    _directory_contents,
    _warn)
from dirq.Exceptions import QueueError, QueueLockError
from dirq.utils import VALID_STR_TYPES, VALID_INT_TYPES

# for backward compatibility.
from dirq.QueueSet import QueueSet

__author__ = dirq.AUTHOR
__version__ = dirq.VERSION
__date__ = dirq.DATE

__all__ = ['Queue', 'QueueSet', 'QueueError']

# name of the directory holding temporary elements
TEMPORARY_DIRECTORY = "temporary"

# name of the directory holding obsolete elements
OBSOLETE_DIRECTORY = "obsolete"

# name of the directory indicating a locked element
LOCKED_DIRECTORY = "locked"

#
# global variables
#

_FILE_REGEXP = re.compile("^([0-9a-zA-Z]+)$")
_KEY_VAL_REGEXP = re.compile("^([^\x09\x0a]*)\x09([^\x09\x0a]*)$")
_H2S_REGEXP = re.compile("(\\\\|\x09|\x0a)")
_S2H_REGEXP = re.compile(r"(\\\\|\\t|\\n)")

_BYTE2ESC = {"\\": r"\\", "\t": r"\t", "\n": r"\n"}
_ESC2BYTE = dict([(_value, _key) for _key, _value in _BYTE2ESC.items()])

#
# Helper Functions
#


def _hash2string(data):
    """Transform a hash of strings into a string.

    Raise:
        QueueError - invalid type of a value in hash
                     (allowed string or unicode)

    Note:
        the keys are sorted so that identical hashes yield to identical strings
    """
    string = ''
    for key in sorted(data.keys()):
        val = data[key]
        if type(val) not in VALID_STR_TYPES:
            raise QueueError("invalid hash value type: %r" % val)
        key = _H2S_REGEXP.sub(lambda m: _BYTE2ESC[m.group(1)], key)
        val = _H2S_REGEXP.sub(lambda m: _BYTE2ESC[m.group(1)], val)
        string = '%s%s' % (string, '%s\x09%s\x0a' % (key, val))
    return string


def _string2hash(given):
    """Transform a string into a hash of strings.

    Raise:
        QueueError - unexpected hash line

    Note:
        duplicate keys are not checked (the last one wins)
    """
    _hash = dict()
    if not given:
        return _hash
    for line in given.strip('\n').split('\x0a'):
        match = _KEY_VAL_REGEXP.match(line)
        if not match:
            raise QueueError("unexpected hash line: %s" % line)
        key = _S2H_REGEXP.sub(lambda m: _ESC2BYTE[str(m.group(1))],
                              match.group(1))
        val = _S2H_REGEXP.sub(lambda m: _ESC2BYTE[str(m.group(1))],
                              match.group(2))
        _hash[key] = val
    return _hash


def _older(path, given_time):
    """
    Check if a path is old enough:

    * return true if the path exists and is (strictly) older than given time
    * return false if it does not exist or it is newer
    * die in case of any other error

    Raise:
        OSError - can't stat given path

    Note:
        lstat() is used so symlinks are not followed
    """
    try:
        stat = os.lstat(path)
    except Exception:
        error = sys.exc_info()[1]
        if error.errno != errno.ENOENT:
            raise OSError("cannot lstat(%s): %s" % (path, error))
            # RACE: this path does not exist (anymore)
        return False
    else:
        return stat.st_mtime < given_time


def __subdirs_num_nlink(path):
    """Count the number of sub-directories in the given directory:

    * return 0 if the directory does not exist (anymore)
    * die in case of any other error

    Raise:
        OSError - can't stat given path

    Note:

    * lstat() is used so symlinks are not followed
    * this only checks the number of links
    * we do not even check that the path indeed points to a directory!
    """
    try:
        stat = os.lstat(path)
    except Exception:
        error = sys.exc_info()[1]
        if error.errno != errno.ENOENT:
            raise OSError("cannot lstat(%s): %s" % (path, error))
            # RACE: this path does not exist (anymore)
        return 0
    else:
        return stat.st_nlink - 2


def __subdirs_num_count(path):
    """Count the number of sub-directories in the given directory:
     - return 0 if the directory does not exist (anymore)

    For systems where we cannot rely on the number of links, so we simply count
    the number of sub-directories.
    """
    return len(_directory_contents(path, missingok=True))


if sys.platform in ['win32', 'cygwin']:
    _subdirs_num = __subdirs_num_count
else:
    _subdirs_num = __subdirs_num_nlink


def _check_element(name):
    """Check the given string to make sure it represents a valid element name.

    Raise:
        QueueError - given element is invalid
    """
    if not _DIRELT_REGEXP.match(name):
        raise QueueError("invalid element name: %s" % name)


def _count(path):
    """Return the number of elements in the queue, regardless of
    their state.

    Raise:
        OSError - can't list/stat element directories
    """
    count = 0
    for name in [x for x in _directory_contents(path)]:
        subdirs = _subdirs_num('%s/%s' % (path, name))
        if subdirs:
            count += subdirs
    return count

#
# Object Oriented Interface
#


class Queue(QueueBase):
    """Directory based queue.
    """
    def __init__(self, path, umask=None, rndhex=None, maxelts=16000,
                 schema=dict()):
        """Check and set schema. Build the queue directory structure.

        Arguments:
            path
                the queue toplevel directory
            umask
                the umask to use when creating files and directories
                (default: use the running process' umask)
            rndhex
                the hexadecimal digit to use in names
                (default: randomly chosen)
            maxelts
                the maximum number of elements that an intermediate
                directory can hold (default: 16,000)
            schema
                the schema defining how to interpret user supplied
                data (mandatory if elements are added or read)
        Raise:
            TypeError  - wrong input data types provided
            QueueError - problems with the queue schema definition
            OSError    - can't create directory structure
        """
        super(Queue, self).__init__(path, umask=umask, rndhex=rndhex)

        if type(maxelts) in VALID_INT_TYPES:
            self.maxelts = maxelts
        else:
            raise TypeError("'maxelts' should be int or long")
        # check schema
        self.type = {}
        self.mandatory = {}
        if schema:
            if not isinstance(schema, dict):
                raise QueueError("invalid schema: %r" % schema)
            for name in schema.keys():
                if not _FILE_REGEXP.match(name):
                    raise QueueError("invalid schema name: %r" % name)
                if not isinstance(schema[name], str):
                    raise QueueError("invalid data type for schema " +
                                     "specification: %r" % type(schema[name]))
                match = re.match('(binary|string|table)([\\?\\*]{0,2})?$',
                                 schema[name])
                if not match:
                    raise QueueError("invalid schema data type: %r" %
                                     schema[name])
                self.type[name] = match.group(1)
                if not re.search('\\?', match.group(2)):
                    self.mandatory[name] = True
            if not self.mandatory:
                raise QueueError("invalid schema: no mandatory data")
        # create directories
        for directory in (TEMPORARY_DIRECTORY, OBSOLETE_DIRECTORY):
            _special_mkdir('%s/%s' % (self.path, directory), self.umask)

    def _is_locked_nlink(self, ename, _time=None):
        """Check if an element is locked.

        Note:
            this is only an indication as the state may be changed by another
            process

        Uses number of links (st_nlink) returned by os.lstat() applied to the
        element directory.

        Arguments:
            ename - name of an element
            _time - consider only locks older than the given time

        Return:
            True  - if element exists and locked. If _time is provided, only
            return True on locks older than this time (needed by purge).

            False - in other cases

        Raises:
            OSError - if unable to stat() the element
        """
        path = '%s/%s' % (self.path, ename)
        try:
            stat = os.lstat(path)
        except Exception:
            error = sys.exc_info()[1]
            if error.errno != errno.ENOENT:
                raise OSError("cannot lstat(%s): %s" % (path, error))
            return False
        # locking increases number of links
        if not stat.st_nlink > 2:
            return False
        # check age if _time is given
        if _time and stat.st_mtime >= _time:
            return False
        return os.path.exists('%s/%s' % (path, LOCKED_DIRECTORY))

    def _is_locked_nonlink(self, ename, _time=None):
        """See _is_locked_nlink(). This version doesn't use nlink (slower).
        """
        path = '%s/%s' % (self.path, ename)
        if not os.path.exists('%s/%s' % (path, LOCKED_DIRECTORY)):
            return False
        elif not _time:
            return True
        # element exists and locked, and we were asked to act upon its age
        try:
            stat = os.lstat(path)
        except Exception:
            error = sys.exc_info()[1]
            if error.errno != errno.ENOENT:
                raise OSError("cannot lstat(%s): %s" % (path, error))
            return False
        return stat.st_mtime < _time

    if sys.platform in ['win32', 'cygwin']:
        _is_locked = _is_locked_nonlink
    else:
        _is_locked = _is_locked_nlink

    def _build_elements(self):
        """Build list of elements.

        Raise:
            OSError - can't list element directories
        """
        while self.dirs:
            directory = self.dirs.pop(0)
            _list = []
            for name in _directory_contents(
                    '%s/%s' % (self.path, directory), True):
                if _ELEMENT_REGEXP.match(name):
                    _list.append(name)
            if not _list:
                continue
            self.elts = ['%s/%s' % (directory, x) for x in sorted(_list)]
            return True
        return False

    def count(self):
        """Return the number of elements in the queue, regardless of
        their state.

        Raise:
            OSError - can't list/stat element directories
        """
        return _count(self.path)

    def lock(self, ename, permissive=True):
        """Lock an element.

        Arguments:
            ename - name of an element
            permissive - work in permissive mode

        Return:

        * True on success
        * False in case the element could not be locked (in permissive
          mode)

        Raise:
            QueueError - invalid element name
            OSError    - can't create lock (mkdir()/lstat() failed)

        Note:

        * locking can fail:

          * if the element has been locked by somebody else (EEXIST)
          * if the element has been removed by somebody else (ENOENT)
        * if the optional second argument is true, it is not an error if
          the element cannot be locked (permissive mode), this is the
          default
        * the directory's mtime will change automatically (after a
          successful mkdir()), this will later be used to detect stalled
          locks
        """
        _check_element(ename)
        path = '%s/%s/%s' % (self.path, ename, LOCKED_DIRECTORY)
        try:
            if self.umask is not None:
                oldumask = os.umask(self.umask)
                os.mkdir(path)
                os.umask(oldumask)
            else:
                os.mkdir(path)
            os.lstat(path)
        except Exception:
            error = sys.exc_info()[1]
            if permissive:
                # RACE: the locked directory already exists
                if error.errno == errno.EEXIST:
                    return False
                # RACE: the element directory does not exist anymore
                if error.errno == errno.ENOENT:
                    return False
            # otherwise this is unexpected...
            raise OSError("cannot mkdir(%s): %s" % (path, error))
        try:
            os.lstat(path)
        except Exception:
            error = sys.exc_info()[1]
            if permissive:
                # RACE: the element directory does not exist anymore (this can
                # happen if an other process locked & removed the element
                # while our mkdir() was in progress... yes, this can happen!)
                if error.errno == errno.ENOENT:
                    return False
            # otherwise this is unexpected...
            raise OSError("cannot lstat(%s): %s" % (path, str(error)))
        return True

    def unlock(self, ename, permissive=False):
        """Unlock an element.

        Arguments:
            ename - name of an element
            permissive - work in permissive mode

        Return:

        * true on success
        * false in case the element could not be unlocked (in permissive
          mode)

        Raise:
            QueueError - invalid element name
            OSError    - can't remove lock (rmdir() failed)

        Note:

        * unlocking can fail:

            * if the element has been unlocked by somebody else (ENOENT)
            * if the element has been removed by somebody else (ENOENT)
        * if the optional second argument is true, it is not an error if
          the element cannot be unlocked (permissive mode), this is _not_
          the default
        """
        _check_element(ename)
        path = '%s/%s/%s' % (self.path, ename, LOCKED_DIRECTORY)
        try:
            os.rmdir(path)
        except Exception:
            error = sys.exc_info()[1]
            if permissive:
                # RACE: the element folder or its lock does not exist anymore
                if error.errno == errno.ENOENT:
                    return False
            raise OSError("cannot rmdir(%s): %s" % (path, error))
        else:
            return True

    def remove(self, ename):
        """Remove locked element from the queue.

        Arguments:
            ename - name of an element

        Raise:
            QueueError - invalid element name; element not locked;
                         unexpected file in the element directory

            OSError    - can't rename/remove a file/directory

        Note:
            doesn't return anything explicitly (i.e. returns NoneType)
            or fails
        """
        _check_element(ename)
        if not self._is_locked(ename):
            raise QueueError("cannot remove %s: not locked" % ename)
        # move the element out of its intermediate directory
        path = '%s/%s' % (self.path, ename)
        while True:
            temp = '%s/%s/%s' % (self.path, OBSOLETE_DIRECTORY,
                                 _name(self.rndhex))
            try:
                os.rename(path, temp)
                break
            except Exception:
                error = sys.exc_info()[1]
                if error.errno != errno.ENOTEMPTY and \
                        error.errno != errno.EEXIST:
                    raise OSError("cannot rename(%s, %s): %s" %
                                  (ename, temp, error))
                    # RACE: the target directory was already present...
        # remove the data files
        for name in _directory_contents(temp):
            if name == LOCKED_DIRECTORY:
                continue
            if not _FILE_REGEXP.match(name):
                raise QueueError("unexpected file in %s: %s" % (temp, name))
            path = '%s/%s' % (temp, name)
            try:
                os.unlink(path)
            except Exception:
                error = sys.exc_info()[1]
                raise OSError("cannot unlink(%s): %s" % (path, error))
        # remove the locked directory
        path = '%s/%s' % (temp, LOCKED_DIRECTORY)
        while True:
            try:
                os.rmdir(path)
            except Exception:
                error = sys.exc_info()[1]
                raise OSError("cannot rmdir(%s): %s" % (path, error))
            try:
                os.rmdir(temp)
                return
            except Exception:
                error = sys.exc_info()[1]
                if error.errno != errno.ENOTEMPTY and \
                        error.errno != errno.EEXIST:
                    raise OSError("cannot rmdir(%s): %s" % (temp, error))
                # RACE: this can happen if an other process managed to lock
                # this element while it was being removed (see the comment in
                # the lock() method) so we try to remove the lock again
                # and again...

    def dequeue(self, ename, permissive=True):
        """Dequeue an element from the queue. Removes element from the
        queue. Performs operations: lock(name), get(name), remove(name)

        Arguments:
            ename - name of an element

        Return:
            dictionary representing an element

        Raise:
            QueueLockError - coulnd't lock element
            QueueError     - problems with schema/data types/etc.
            OSError        - problems opening/closing directory/file
        """
        if not self.lock(ename, permissive=permissive):
            raise QueueLockError("couldn't lock element: %s" % ename)
        element = self.get(ename)
        self.remove(ename)
        return element

    def get(self, ename):
        """Get an element data from a locked element.

        Arguments:
            ename - name of an element

        Return:
            dictionary representing an element

        Raise:
            QueueError - schema is unknown; unexpected data type in
                         the schema specification; missing mandatory
                         file of the element

            OSError    - problems opening/closing file
            IOError    - file read error
        """
        if not self.type:
            raise QueueError("unknown schema")
        _check_element(ename)
        if not self._is_locked(ename):
            raise QueueError("cannot get %s: not locked" % ename)
        data = {}
        for dname in self.type.keys():
            path = '%s/%s/%s' % (self.path, ename, dname)
            try:
                os.lstat(path)
            except Exception:
                error = sys.exc_info()[1]
                if error.errno != errno.ENOENT:
                    raise OSError("cannot lstat(%s): %s" % (path, error))
                if dname in self.mandatory:
                    raise QueueError("missing data file: %s" % path)
                else:
                    continue
            if self.type[dname] == 'binary':
                data[dname] = _file_read(path, 0)
            elif self.type[dname] == 'string':
                data[dname] = _file_read(path, 1)
            elif self.type[dname] == 'table':
                data[dname] = _string2hash(_file_read(path, 1))
            else:
                raise QueueError("unexpected data type: %s" % self.type[dname])
        return data

    def get_element(self, ename, permissive=True):
        """Get an element from the queue. Element will not be removed.
        Operations performed: lock(name), get(name), unlock(name)

        Arguments:
            ename - name of an element

        Raise:
            QueueLockError - couldn't lock element
        """
        if not self.lock(ename, permissive=permissive):
            raise QueueLockError("couldn't lock element: %s" % ename)
        element = self.get(ename)
        self.unlock(ename, permissive=permissive)
        return element

    def _insertion_directory(self):
        """Return the name of the intermediate directory that can be used for
        insertion:

        * if there is none, an initial one will be created
        * if it is full, a new one will be created
        * in any case the name will match $_DIRECTORY_REGEXP

        Raise:
            OSError - can't list/make element directories
        """
        _list = []
        # get the list of existing directories
        for name in _directory_contents(self.path):
            if _DIRECTORY_REGEXP.match(name):
                _list.append(name)
        # handle the case with no directories yet
        if not _list:
            name = '%08x' % 0
            _special_mkdir('%s/%s' % (self.path, name), self.umask)
            return name
        # check the last directory
        _list.sort()
        name = _list[-1]
        subdirs = _subdirs_num('%s/%s' % (self.path, name))
        if subdirs:
            if subdirs < self.maxelts:
                return name
        else:
            # RACE: at this point, the directory does not exist anymore,
            # so it must have been purged after we listed the directory
            # contents. We do not try to do more and simply create a new
            # directory
            pass
        # we need a new directory
        name = '%08x' % (int(name, 16) + 1)
        _special_mkdir('%s/%s' % (self.path, name), self.umask)
        return name

    def add(self, data):
        """Add a new element to the queue and return its name.
        Arguments:

            data - element as a dictionary (should conform to the schema)

        Raise:

            QueueError - problem with schema definition or data
            OSError    - problem putting element on disk

        Note:

          the destination directory must _not_ be created beforehand as
          it would be seen as a valid (but empty) element directory by
          another process, we therefore use rename() from a temporary
          directory
        """
        if not self.type:
            raise QueueError("unknown schema")
        while True:
            temp = '%s/%s/%s' % (self.path, TEMPORARY_DIRECTORY,
                                 _name(self.rndhex))
            if _special_mkdir(temp, self.umask):
                break
        for name in data.keys():
            if name not in self.type:
                raise QueueError("unexpected data: %s" % name)
            if self.type[name] == 'binary':
                if type(data[name]) not in VALID_STR_TYPES:
                    raise QueueError("unexpected binary data in %s: %r" %
                                     (name, data[name]))
                _file_write('%s/%s' % (temp, name), 0, self.umask, data[name])
            elif self.type[name] == 'string':
                if type(data[name]) not in VALID_STR_TYPES:
                    raise QueueError("unexpected string data in %s: %r" %
                                     (name, data[name]))
                _file_write('%s/%s' % (temp, name), 1, self.umask, data[name])
            elif self.type[name] == 'table':
                if not isinstance(data[name], dict):
                    raise QueueError("unexpected table data in %s: %r" %
                                     (name, data[name]))
                _file_write('%s/%s' % (temp, name), 1, self.umask,
                            _hash2string(data[name]))
            else:
                raise QueueError("unexpected data type in %s: %r" %
                                 (name, self.type[name]))
        for name in self.mandatory.keys():
            if name not in data:
                raise QueueError("missing mandatory data: %s" % name)
        while True:
            name = '%s/%s' % (self._insertion_directory(), _name(self.rndhex))
            path = '%s/%s' % (self.path, name)
            try:
                os.rename(temp, path)
                return name
            except Exception:
                error = sys.exc_info()[1]
                if error.errno != errno.ENOTEMPTY and \
                        error.errno != errno.EEXIST:
                    raise OSError("cannot rename(%s, %s): %s" %
                                  (temp, path, error))
                    # RACE: the target directory was already present...
    enqueue = add

    def _volatile(self):
        """
        Return the list of volatile (i.e. temporary or obsolete) directories.
        """
        _list = []
        for name in _directory_contents(
                '%s/%s' % (self.path, TEMPORARY_DIRECTORY), True):
            if _ELEMENT_REGEXP.match(name):
                _list.append('%s/%s' % (TEMPORARY_DIRECTORY, name))
        for name in _directory_contents(
                '%s/%s' % (self.path, OBSOLETE_DIRECTORY), True):
            if _ELEMENT_REGEXP.match(name):
                _list.append('%s/%s' % (OBSOLETE_DIRECTORY, name))
        return _list

    def purge(self, maxtemp=300, maxlock=600):
        """Purge the queue:

        * delete unused intermediate directories
        * delete too old temporary directories
        * unlock too old locked directories

        Arguments:
            maxtemp - maximum time for a temporary element. If 0, temporary
                      elements will not be removed.
            maxlock - maximum time for a locked element. If 0, locked
                      elements will not be unlocked.
        Raise:
            OSError - problem deleting element from disk

        Note:
            this uses first()/next() to iterate so this will reset the cursor
        """
        # get the list of intermediate directories
        _list = []
        for name in _directory_contents(self.path):
            if _DIRECTORY_REGEXP.match(name):
                _list.append(name)
        _list.sort()
        # try to purge all but last one
        if len(_list) > 1:
            _list.pop()
            for name in _list:
                path = '%s/%s' % (self.path, name)
                if _subdirs_num(path):
                    continue
                _special_rmdir(path)
        # remove the volatile directories which are too old
        if maxtemp:
            oldtime = time.time() - maxtemp
            for name in self._volatile():
                path = '%s/%s' % (self.path, name)
                if _older(path, oldtime):
                    _warn("* removing too old volatile element: %s" % name)
                    for file_name in _directory_contents(path, True):
                        if file_name == LOCKED_DIRECTORY:
                            continue
                        fpath = '%s/%s' % (path, file_name)
                        try:
                            os.unlink(fpath)
                        except Exception:
                            error = sys.exc_info()[1]
                            if error.errno != errno.ENOENT:
                                raise OSError("cannot unlink(%s): %s" %
                                              (fpath, error))
                _special_rmdir('%s/%s' % (path, LOCKED_DIRECTORY))
                _special_rmdir(path)
        # iterate to find abandoned locked entries
        if maxlock:
            oldtime = time.time() - maxlock
            name = self.first()
            while name:
                if self._is_locked(name, oldtime):
                    _warn("* removing too old locked element: %s" % name)
                    self.unlock(name, True)
                name = self.next()



"""
QueueSimple - object oriented interface to a simple directory based queue.

A port of Perl module Directory::Queue::Simple
http://search.cpan.org/dist/Directory-Queue/
The documentation from Directory::Queue::Simple module was
adapted for Python.

=================
QueueSimple class
=================

:py:class:`QueueSimple` - simple directory based queue.

Usage::

    from dirq.QueueSimple import QueueSimple

    # sample producer

    dirq = QueueSimple('/tmp/test')
    for count in range(1,101):
        name = dirq.add("element %i\\n" % count)
        print("# added element %i as %s" % (count, name))

    # sample consumer

    dirq = QueueSimple('/tmp/test')
    for name in dirq:
        if not dirq.lock(name):
            continue
        print("# reading element %s" % name)
        data = dirq.get(name)
        # one could use dirq.unlock(name) to only browse the queue...
        dirq.remove(name)


Description
-----------

    This module is very similar to dirq.queue, but uses a
    different way to store data in the filesystem, using less
    directories. Its API is almost identical.

    Compared to :py:mod:`dirq.queue`, this module:

    * is simpler
    * is faster
    * uses less space on disk
    * can be given existing files to store
    * does not support schemas
    * can only store and retrieve byte strings
    * is not compatible (at filesystem level) with Queue

    Please refer to :py:mod:`dirq.queue` for general information about
    directory queues.

Directory Structure
-------------------

    The toplevel directory contains intermediate directories that contain
    the stored elements, each of them in a file.

    The names of the intermediate directories are time based: the element
    insertion time is used to create a 8-digits long hexadecimal number.
    The granularity (see the constructor) is used to limit the number of
    new directories. For instance, with a granularity of 60 (the default),
    new directories will be created at most once per minute.

    Since there is usually a filesystem limit in the number of directories
    a directory can hold, there is a trade-off to be made. If you want to
    support many added elements per second, you should use a low
    granularity to keep small directories. However, in this case, you will
    create many directories and this will limit the total number of
    elements you can store.

    The elements themselves are stored in files (one per element) with a
    14-digits long hexadecimal name SSSSSSSSMMMMMR where:

    * SSSSSSSS represents the number of seconds since the Epoch

    * MMMMM represents the microsecond part of the time since the Epoch

    * R is a random digit used to reduce name collisions


    A temporary element (being added to the queue) will have a *.tmp*
    suffix.

    A locked element will have a hard link with the same name and the
    *.lck* suffix.


Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import errno
import os
import re
import sys
import time

from dirq.QueueBase import QueueBase, _name, _file_create, _special_mkdir, \
    _file_read, _DIRECTORY_REGEXP, _ELEMENT_REGEXP, _special_rmdir, _warn
from dirq.utils import is_bytes

# suffix indicating a temporary element
TEMPORARY_SUFFIX = ".tmp"

# suffix indicating a locked element
LOCKED_SUFFIX = ".lck"


class QueueSimple(QueueBase):
    """
    QueueSimple
    """
    def __init__(self, path, umask=None, rndhex=None, granularity=60):
        """
        * path - queue top level directory
        * umask - the umask to use when creating files and directories
                  (default: use the running process' umask)
        * rndhex - the hexadecimal digit to use in names
                  (default: randomly chosen)
        * granularity - the time granularity for intermediate directories
                      (default: 60)
        """
        super(QueueSimple, self).__init__(path, umask=umask, rndhex=rndhex)

        if not isinstance(granularity, int):
            raise TypeError('granularity should be integer.')
        self.granularity = granularity

    def _add_dir(self):
        """ Return new directory name based on time and granularity. """
        now = int(time.time())
        if self.granularity > 1:
            now -= now % self.granularity
        return "%08x" % now

    def _add_data(self, data):
        """Write 'data' to a file.

        Return: (tuple) directory name where the file was written, full path to
        the temporary file.
        """
        _dir = self._add_dir()
        while 1:
            tmp = '%s/%s/%s%s' % (self.path, _dir, _name(self.rndhex),
                                  TEMPORARY_SUFFIX)
            try:
                if is_bytes(data):
                    new_file = _file_create(tmp, umask=self.umask, utf8=False)
                else:
                    new_file = _file_create(tmp, umask=self.umask, utf8=True)
            except EnvironmentError:
                error = sys.exc_info()[1]
                if error.errno == errno.ENOENT:
                    _special_mkdir('%s/%s' % (self.path, _dir), self.umask)
                    continue
            else:
                if new_file:
                    break
        new_file.write(data)
        new_file.close()
        return _dir, tmp

    def _add_path(self, tmp, _dir):
        """Given temporary file and directory where it resides: create a hard
        link to that file and remove initial one.

        Return: element name (<directory name>/<file name>).
        """
        while 1:
            name = _name(self.rndhex)
            new = '%s/%s/%s' % (self.path, _dir, name)
            try:
                os.link(tmp, new)
            except OSError:
                error = sys.exc_info()[1]
                if error.errno != errno.EEXIST:
                    raise error
                else:
                    continue
            os.unlink(tmp)
            return '%s/%s' % (_dir, name)

    def _build_elements(self):
        _list = []
        while self.dirs:
            _dir = self.dirs.pop(0)
            for name in os.listdir('%s/%s' % (self.path, _dir)):
                if _ELEMENT_REGEXP.match(name):
                    _list.append(name)
            if not _list:
                continue
            _list.sort()
            self.elts = ['%s/%s' % (_dir, x) for x in _list]
            return True
        return False

    def add(self, data):
        """Add data to the queue as a file.

        Return: element name (<directory name>/<file name>).
        """
        _dir, path = self._add_data(data)
        return self._add_path(path, _dir)

    add_ref = add  # to comply with the Perl Directory::Queue interface

    def add_path(self, path):
        """Add the given file (identified by its path) to the queue and return
        the corresponding element name, the file must be on the same
        filesystem and will be moved to the queue
        """
        _dir = self._add_dir()
        _special_mkdir('%s/%s' % (self.path, _dir), self.umask)
        return self._add_path(path, _dir)

    def get(self, name):
        """Get locked element.
        """
        return _file_read('%s/%s%s' % (self.path, name, LOCKED_SUFFIX), False)

    get_ref = get  # to comply with the Perl Directory::Queue interface

    def get_path(self, name):
        """ Return the path given the name. """
        return '%s/%s%s' % (self.path, name, LOCKED_SUFFIX)

    def lock(self, name, permissive=True):
        """Lock an element.

        Arguments:
            name - name of an element
            permissive - work in permissive mode

        Return:

        * true on success
        * false in case the element could not be locked (in permissive
          mode)
        """
        path = '%s/%s' % (self.path, name)
        lock = '%s%s' % (path, LOCKED_SUFFIX)
        try:
            os.link(path, lock)
        except OSError:
            error = sys.exc_info()[1]
            if permissive and (error.errno == errno.EEXIST or
                               error.errno == errno.ENOENT):
                return False
            new_error = OSError("cannot link(%s, %s): %s" %
                                (path, lock, error))
            new_error.errno = error.errno
            raise new_error
        else:
            try:
                os.utime(path, None)
            except OSError:
                # RACE: the element file does not exist anymore
                # (this can happen if an other process locked & removed the
                # element while our link() was in progress...
                # yes, this can happen!)
                error = sys.exc_info()[1]
                if permissive and error.errno == errno.ENOENT:
                    os.unlink(lock)
                    return False
                new_error = OSError("cannot utime(%s, %s): %s" %
                                    (path, lock, error))
                new_error.errno = error.errno
                raise new_error
            else:
                return True

    def unlock(self, name, permissive=False):
        """Unlock an element.

        Arguments:
            name - name of an element
            permissive - work in permissive mode

        Return:

        * true on success
        * false in case the element could not be unlocked (in permissive
          mode)
        """
        lock = '%s/%s%s' % (self.path, name, LOCKED_SUFFIX)
        try:
            os.unlink(lock)
        except OSError:
            error = sys.exc_info()[1]
            if permissive and error.errno == errno.ENOENT:
                return False
            raise error
        else:
            return True

    def remove(self, name):
        """Remove a locked element from the queue.
        """
        path = '%s/%s' % (self.path, name)
        lock = '%s%s' % (path, LOCKED_SUFFIX)
        os.unlink(path)
        os.unlink(lock)

    def __get_list_of_interm_dirs(self, dirs):
        """Fill out provided list with names of intermediate directories.
        """
        for name in os.listdir(self.path):
            if _DIRECTORY_REGEXP.match(name):
                dirs.append(name)

    def count(self):
        """Return the number of elements in the queue, locked or not
        (but not temporary).
        """
        count = 0
        # get list of intermediate directories
        dirs = []
        self.__get_list_of_interm_dirs(dirs)
        # count elements in sub-directories
        for name in dirs:
            for element in os.listdir('%s/%s' % (self.path, name)):
                if _ELEMENT_REGEXP.match(element):
                    count += 1
        return count

    def purge(self, maxtemp=300, maxlock=600):
        """Purge the queue by removing unused intermediate directories,
        removing too old temporary elements and unlocking too old locked
        elements (aka staled locks); note: this can take a long time on
        queues with many elements.

        maxtemp - maximum time for a temporary element
                  (in seconds, default 300);
                  if set to 0, temporary elements will not be removed
        maxlock - maximum time for a locked element
                  (in seconds, default 600);
                  if set to 0, locked elements will not be unlocked
        """
        # get list of intermediate directories
        dirs = []
        self.__get_list_of_interm_dirs(dirs)
        # remove all but old temporary or locked elements
        oldtemp = maxtemp != 0 and time.time() - maxtemp or 0
        oldlock = maxlock != 0 and time.time() - maxlock or 0
        if oldtemp or oldlock:
            for _dir in dirs:
                path = '%s/%s' % (self.path, _dir)
                tmp_lock_elems = [x for x in os.listdir(path)
                                  if re.search('(%s|%s)$' %
                                               (TEMPORARY_SUFFIX,
                                                LOCKED_SUFFIX), x)]
                for old in tmp_lock_elems:
                    try:
                        stat = os.stat('%s/%s' % (path, old))
                    except OSError:
                        error = sys.exc_info()[1]
                        if error.errno == errno.ENOENT:
                            continue
                        raise error
                    if (old.endswith(TEMPORARY_SUFFIX) and
                            stat.st_mtime >= oldtemp):
                        continue
                    if old.endswith(LOCKED_SUFFIX) and \
                            stat.st_mtime >= oldlock:
                        continue
                    _warn("WARNING: removing too old volatile file: %s/%s" %
                          (self.path, old))
                    try:
                        os.unlink('%s/%s' % (path, old))
                    except OSError:
                        error = sys.exc_info()[1]
                        if error.errno != errno.ENOENT:
                            raise error
        # try to purge all but the last intermediate directory
        if len(dirs) > 1:
            dirs.sort()
            dirs.pop()
            for _dir in dirs:
                path = '%s/%s' % (self.path, _dir)
                if len(os.listdir(path)) == 0:
                    _special_rmdir(path)



# -*- coding: utf-8 -*-

import codecs
import errno
import os
import shutil
import sys
import tempfile
import unittest

from dirq import QueueBase
from dirq.QueueBase import QueueBase as QueueBaseClass

__all__ = ['TestQueueBase', 'TestQueueBaseModuleFunctions']


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq-base')
        self.qdir = self.tempdir + '/dirq'
        self.tempfile = self.tempdir + '/file'

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueueBase(TestDirQueue):

    def test1init(self):
        'QueueBase.__init__()'
        path = self.tempdir + '/aaa/bbb/ccc'
        QueueBaseClass(path, umask=None)
        assert os.path.exists(path) is True


class TestQueueBaseModuleFunctions(TestDirQueue):

    def test1_special_mkdir(self):
        'QueueBase._special_mkdir()'
        assert QueueBase._special_mkdir(self.qdir) is True
        assert QueueBase._special_mkdir(self.qdir) is False
        # test against a file
        test_file = os.path.join(self.qdir, 'foo')
        open(test_file, 'w').write('bar')
        self.assertRaises(OSError, QueueBase._special_mkdir, (test_file))

    def test2_name(self):
        'QueueBase._name()'
        n = QueueBase._name(7)
        assert len(n) == 14
        assert n.endswith('%01x' % 7)

    def test3_file_create(self):
        'QueueBase._file_create()'
        # File in non existent directory should produce ENOENT.
        fn = os.getcwd() + "/nodir/nofile"
        try:
            QueueBase._file_create(fn, 0, False)
        except OSError:
            error = sys.exc_info()[1]
            assert error.errno == errno.ENOENT

        QueueBase._file_create(self.tempfile, 0, False)
        self.assertRaises(OSError,
                          QueueBase._file_create,
                          *(self.tempfile, 0, False))
        os.unlink(self.tempfile)
        # utf8 data
        QueueBase._file_create(self.tempfile, 0, True)
        self.assertRaises(OSError,
                          QueueBase._file_create,
                          *(self.tempfile, 0, True))

    def test4_file_write(self):
        'QueueBase._file_write()'
        QueueBase._file_write(self.tempfile, 0, False, 'a\n'.encode())
        os.unlink(self.tempfile)
        QueueBase._file_write(self.tempfile, 0,
                              False, ('a' * (2 ** 10) * 10).encode())
        os.unlink(self.tempfile)
        for t in [1, [], (), {}, object]:
            self.assertRaises(TypeError, QueueBase._file_write, ('', t))

    def test5_file_read(self):
        'QueueBase._file_read()'
        text = 'hello\n'.encode()
        open(self.tempfile, 'wb').write(text)
        text_in = QueueBase._file_read(self.tempfile, False)
        self.assertEqual(text, text_in)
        # utf8
        try:
            text = 'Élément \u263A\n'.decode("utf-8")
        except AttributeError:
            text = 'Élément \u263A\n'
        codecs.open(self.tempfile, 'w', 'utf8').write(text)
        text_in = QueueBase._file_read(self.tempfile, True)
        self.assertEqual(text, text_in)


def main():
    testcases = [TestQueueBase,
                 TestQueueBaseModuleFunctions]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



"""
Utilities used in dirq module.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import sys

_PY2 = sys.hexversion < 0x03000000
_PY3 = not _PY2

if _PY2:
    VALID_STR_TYPES = (str, unicode)
else:
    VALID_STR_TYPES = (str, bytes)

if _PY2:
    VALID_INT_TYPES = (int, long)
else:
    VALID_INT_TYPES = (int, )


def is_bytes(string):
    """ Check if given string is a byte string. """
    if _PY2:
        return not isinstance(string, unicode)
    else:  # python 3
        return isinstance(string, bytes)



# -*- coding: utf-8 -*-

import inspect
import os
import shutil
import subprocess
import sys
import tempfile
import unittest

__all__ = ['TestQueueWrapper']


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq-wrapper')
        self.qdir = self.tempdir + '/dirq'
        me = inspect.getfile(inspect.currentframe())
        self.tdir = os.path.dirname(os.path.abspath(me))
        self.args = ["-p", self.qdir]
        if __name__ == "__main__":
            self.args.append("-d")

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueueWrapper(TestDirQueue):

    def test01(self):
        'dqt.py normal'
        tcmd = [sys.executable, self.tdir + "/dqt.py"] + self.args
        rc = subprocess.call(tcmd +
                             ["-c", "1000", "--type", "normal", "simple"])
        assert rc == 0

    def test02(self):
        'dqt.py simple'
        tcmd = [sys.executable, self.tdir + "/dqt.py"] + self.args
        rc = subprocess.call(tcmd +
                             ["-c", "1000", "--type", "simple", "simple"])
        assert rc == 0

    def test03(self):
        'dqst.py complex'
        tcmd = [sys.executable, self.tdir + "/dqst.py"] + self.args
        rc = subprocess.call(tcmd +
                             ["-c", "200", "--type", "normal", "complex"])
        assert rc == 0

    def test04(self):
        'dqst.py iterate'
        tcmd = [sys.executable, self.tdir + "/dqt.py",
                "-c", "1000", "--type", "simple"]
        if __name__ == "__main__":
            tcmd.append("-d")
        rc = subprocess.call(tcmd + ["-p", self.tempdir + "/q1", "add"])
        assert rc == 0
        rc = subprocess.call(tcmd + ["-p", self.tempdir + "/q2", "add"])
        assert rc == 0
        tcmd = [sys.executable, self.tdir + "/dqst.py", "--type", "simple"]
        if __name__ == "__main__":
            tcmd.append("-d")
        paths = "%s/q1,%s/q2" % (self.tempdir, self.tempdir)
        rc = subprocess.call(tcmd + ["-p", paths, "iterate"])
        assert rc == 0


def main():
    testcases = [TestQueueWrapper]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



"""
QueueNull - object oriented interface to a null directory based queue.

A port of Perl module Directory::Queue::Null
http://search.cpan.org/dist/Directory-Queue/
The documentation from Directory::Queue::Null module was
adapted for Python.

=================
QueueNull class
=================

:py:class:`QueueNull` - null directory based queue.

Usage::

    from dirq.QueueNull import QueueNull

    # sample producer

    dirq = QueueNull()
    for count in range(1,101):
        name = dirq.add("element %i\\n" % count)


Description
-----------

    The goal of this module is to offer a "null" queue system using the
    same API as the other directory queue implementations. The queue will
    behave like a black hole: added data will disappear immediately so the
    queue will therefore always appear empty.

    This can be used for testing purposes or to discard data like one
    would do on Unix by redirecting output to */dev/null*.

    Please refer to :py:mod:`dirq.queue` for general information about
    directory queues.


Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import os
from dirq.QueueBase import QueueBase


class QueueNull(QueueBase):
    """
    QueueNull
    """
    def __init__(self):
        """ Constructor, this does nothing."""
        pass

    def add(self, data):
        """Add data to the queue, this does nothing. """
        return ""

    add_ref = add  # to comply with the Perl Directory::Queue interface

    def add_path(self, path):
        """Add the given file (identified by its path) to the queue,
        this will therefore *remove the file.*
        """
        os.unlink(path)
        return ""

    def get(self, name):
        """ Not supported method. """
        raise NotImplementedError("unsupported method: get()")

    get_ref = get  # to comply with the Perl Directory::Queue interface

    def get_path(self, name):
        """ Not supported method. """
        raise NotImplementedError("unsupported method: get_path()")

    def lock(self, name, permissive=True):
        """ Not supported method. """
        raise NotImplementedError("unsupported method: lock()")

    def unlock(self, name, permissive=False):
        """ Not supported method. """
        raise NotImplementedError("unsupported method: unlock()")

    def remove(self, name):
        """ Not supported method. """
        raise NotImplementedError("unsupported method: remove()")

    def count(self):
        """Return the number of elements in the queue, which means
        it always return 0.
        """
        return 0

    def purge(self, maxtemp=300, maxlock=600):
        """ Purge the queue, this does nothing. """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Browse a set of queues. """

from dirq.queue import Queue, QueueSet
import tempfile

# root directory for queues
working_dir = tempfile.mkdtemp()
# number of queues
QUEUES = 4

print("*** Setup & populate queues")
# generate paths
paths = []
for i in range(QUEUES):
    paths.append(working_dir + '/q-%i' % i)
COUNT = 5

print("creating %i initial queues. adding %i elements into each." % (QUEUES,
                                                                     COUNT))
queues = []
queue_num = 0
while queue_num < QUEUES:
    queue = Queue(paths[queue_num], maxelts=5, schema={'body': 'string'})
    print("adding %d elements to the queue %s" % (COUNT, paths[queue_num]))
    element = {}
    done = 0
    while not COUNT or done < COUNT:
        done += 1
        element['body'] = 'Queue %i. Element %i' % (queue_num, done)
        queue.add(element)
    queues.append(queue)
    queue_num += 1
print("done.")


print("*** Browse")
i = 2
queue_set = QueueSet(queues[0:i])
queue_set_count = queue_set.count()
print("elements in %i queues: %i" % (i, queue_set_count))
assert queue_set_count == i * COUNT

print("adding remaining queues to the set.")
queue_set.add(queues[i:])
total_inset = queue_set.count()
print("total element with added queues: %d" % total_inset)
assert total_inset == QUEUES * COUNT

print("removing %i first queues." % i)
for q in queues[0:i]:
    queue_set.remove(q)

total_inset = queue_set.count()
print("number of elements left in the set: %d" % total_inset)
assert total_inset == (QUEUES - i) * COUNT

print("iterating over the elements left in the queue set")
for q, name in queue_set:
    print(q.path, name)
    print(q.get_element(name)['body'])
print("done.")



"""Interface to elements on a set of directory based queues.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""

import dirq

from dirq.QueueBase import QueueBase
from dirq.Exceptions import QueueError

__version__ = dirq.VERSION
__author__ = dirq.AUTHOR
__date__ = dirq.DATE


class QueueSet(object):
    """Interface to elements on a set of directory based queues.
    """
    def __init__(self, *queues):
        """Generate queue set on the given lists of queues. Copies of the
        object instances are used.

        Arguments:
            *queues - QueueSet([q1,..]/(q1,..)) or QueueSet(q1,..)

        Raise:
            QueueError - queues should be list/tuple or Queue object
            TypeError  - one of objects provided is not instance of Queue
        """
        self.qset = []  # set of queues
        self.elts = []  # local (queue, element) cache
        self._next_exception = False

        self._add(*queues)

    def __iter__(self):
        """ Return iterator over element names on the set of queues. """
        self._reset()
        self._next_exception = True
        return self

    def names(self):
        """Return iterator over element names on the set of queues.
        """
        return self.__iter__()

    def _reset(self):
        """Regenerate lists of intermediate directories and drop cached
        elements lists.

        Raise:
            OSError - can't list directories
        """
        for queue in self.qset:
            queue._reset()
        self.elts = []

    def first(self):
        """Return the first element in the queue set and cache information
        about the next ones.

        Raise:
            OSError - can't list directories
        """
        self._reset()
        return self.next()

    def __next__(self):
        """Return (queue, next element) tuple from the queue set, only using
        cached information.

        Raise:
            StopIteration - when used as Python iterator via
                            __iter__() method

            OSError       - can't list element directories
        """
        if not self.elts:
            for queue in self.qset:
                self.elts.append((queue, queue.next()))
            if not self.elts:
                return (None, None)
        self.elts.sort(key=lambda x: x[1])
        for index, queue_elt in enumerate(self.elts):
            self.elts[index] = (queue_elt[0], queue_elt[0].next())
            if queue_elt[1]:
                return queue_elt
        if self._next_exception:
            self._next_exception = False
            raise StopIteration
        else:
            return (None, None)
    next = __next__

    def count(self):
        """Return the number of elements in the queue set, regardless of
        their state.

        Raise:
            OSError - can't list/stat element directories
        """
        count = 0
        for queue in self.qset:
            count += queue.count()
        return count

    def _add(self, *queues):
        """Add lists of queues to existing ones. Copies of the object
        instances are used.

        Arguments:
            *queues - add([q1,..]/(q1,..)) or add(q1,..)

        Raise:
            QueueError - queue already in the set
            TypeError  - wrong queue object type provided
        """
        type_queue = False
        for queue in queues:
            if type(queue) in [list, tuple] and not type_queue:
                for _queue in queue:
                    if isinstance(_queue, QueueBase):
                        if _queue.id in [x.id for x in self.qset]:
                            raise QueueError("queue already in the set: %s" %
                                             _queue.path)
                        self.qset.append(_queue.copy())
                    else:
                        raise TypeError("QueueBase objects expected.")
                break
            elif isinstance(queue, QueueBase):
                type_queue = True
                self.qset.append(queue.copy())
            else:
                raise TypeError("expected QueueBase object(s) or list/tuple "
                                "of QueueBase objects")

    def add(self, *queues):
        """Add lists of queues to existing ones. Copies of the object
        instances are used.

        Arguments:
            *queues - add([q1,..]/(q1,..)) or add(q1,..)

        Raise:
            QueueError - queue already in the set
            TypeError  - wrong queue object type provided
        """
        self._add(*queues)
        self._reset()

    def remove(self, given_queue):
        """Remove a queue and its respective elements from in memory cache.

        Arguments:
            queue - queue to be removed

        Raise:
            TypeError - wrong queue object type provided
        """
        if not isinstance(given_queue, QueueBase):
            raise TypeError("QueueBase objects expected.")
        for index, queue in enumerate(self.qset):
            if given_queue.id == queue.id:
                del self.qset[index]
                if self.elts:
                    del self.elts[index]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test program for testing the dirq.queue and dirq.QueueSimple modules.
"""

import os
import random
import re
import shutil
import sys
import time

from optparse import OptionParser

sys.path.insert(1, re.sub(r'/\w*$', '', os.getcwd()))

import dirq  # noqa E402
from dirq import queue  # noqa E402
from dirq.QueueSimple import QueueSimple  # noqa E402

opts = None
TEST = ''
TESTS = ['all', 'add', 'count', 'get', 'iterate', 'purge', 'remove', 'simple']
ProgramName = sys.argv[0]


def init():
    """ Initialize. """
    global opts, TEST
    parser = OptionParser(usage="%prog [OPTIONS] [--] TEST",
                          version="%%prog %s" % dirq.VERSION)
    parser.add_option("-l", "--list", dest="list", action="store_true",
                      default=False, help="list available tests")
    parser.add_option("-d", "--debug", dest="debug", action="store_true",
                      default=False, help="show debugging information")
    parser.add_option("-p", "--path", dest="path", type="string", default="",
                      help="set the queue path")
    parser.add_option("-c", "--count", dest="count", type="int", default=0,
                      help="set the elements count")
    parser.add_option("-s", "--size", dest="size", type="int", default=0,
                      help="set the body size for added elements")
    parser.add_option("-r", "--random", dest="random", action="store_true",
                      default=False, help="randomize the body size")
    parser.add_option("--granularity", dest="granularity", type="int",
                      default=None, help="time granularity for intermediate "
                      "directories (QueueSimple)")
    parser.add_option("--header", dest="header", action="store_true",
                      default=False, help="set header for added elements")
    parser.add_option("--maxelts", dest="maxelts", type="int",
                      default=0, help="set the maximum number of elements per "
                      "directory")
    parser.add_option("--maxlock", dest="maxlock", type="int", default=None,
                      help="maximum time for a locked element. 0 - locked "
                      "elements will not be unlocked.")
    parser.add_option("--maxtemp", dest="maxtemp", type="int", default=None,
                      help="maximum time for a temporary element. "
                      "0 - temporary elements will not be removed.")
    parser.add_option("--rndhex", dest="rndhex", type="int",
                      default=None, help="set the queue random hex digit")
    parser.add_option("--sleep", dest="sleep", type="float", default=0,
                      help="sleep this amount of seconds before starting "
                           "the test(s)")
    parser.add_option("--type", dest="type", type="string", default="simple",
                      help="set the type of dirq (simple|normal)")
    opts, args = parser.parse_args()
    if opts.list:
        print("Tests: %s" % ', '.join(TESTS))
        sys.exit()
    if not opts.path:
        _die("%s: mandatory option not set: -p/--path", ProgramName)
    if len(args) != 0:
        TEST = args[0]
        if TEST not in TESTS:
            _die("Unsupported test '%s'.\nTEST should be one of: %s" %
                 (TEST, ', '.join(TESTS)))
    else:
        parser.print_help()
        sys.exit()


def _die(fmt, *arguments):
    """Report a fatal error."""
    sys.stderr.write(fmt % arguments + "\n")
    sys.stderr.flush()
    sys.exit(1)


def debug(fmt, *arguments):
    """Report a debugging message.
    """
    if not opts.debug:
        return
    message = fmt % arguments
    message = re.sub(r'\s+$', '.', message)
    sys.stderr.write("# %s %s[%d]: %s\n" %
                     (time.strftime("%Y/%m/%d-%H:%M:%S",
                                    time.localtime(time.time())),
                      os.path.basename(sys.argv[0]),
                      os.getpid(), message))


def new_dirq(_schema):
    """Create a new Directory::Queue object, optionally with schema.
    """
    kwargs = {}
    if opts.type == "simple":
        if opts.granularity is not None:
            kwargs['granularity'] = opts.granularity
        if opts.rndhex is not None:
            kwargs['rndhex'] = opts.rndhex
        return QueueSimple(opts.path, **kwargs)
    else:
        if _schema:
            schema = {'body': 'string',
                      'header': 'table?'}
            kwargs['schema'] = schema
        if opts.maxelts:
            kwargs['maxelts'] = opts.maxelts
        if opts.rndhex is not None:
            kwargs['rndhex'] = opts.rndhex
        return queue.Queue(opts.path, **kwargs)


def test_count():
    """Count the elements in the queue.
    """
    dq = new_dirq(0)
    time1 = time.time()
    count = dq.count()
    time2 = time.time()
    debug("queue has %d elements", count)
    debug("done in %.4f seconds", time2 - time1)


def test_purge():
    """Purge the queue.
    """
    debug("purging the queue...")
    dq = new_dirq(0)
    time1 = time.time()
    pwkargs = {}
    if opts.maxtemp is not None:
        pwkargs['maxtemp'] = opts.maxtemp
    if opts.maxlock is not None:
        pwkargs['maxlock'] = opts.maxlock
    dq.purge(**pwkargs)
    time2 = time.time()
    debug("done in %.4f seconds", time2 - time1)


def _body(size, rand):
    """Return a test body"""
    if rand:
        # see Irwin-Hall in http://en.wikipedia.org/wiki/Normal_distribution
        rnd = 0.
        for _ in range(12):
            rnd += random.random()
        rnd -= 6.
        rnd *= size / 6
        size += int(rnd)
    if size < 1:
        return ''
    return "A" * (size - 1) + "\n"


def test_add():
    """Add elements to the queue.
    """
    rnd = opts.random
    size = opts.size
    count = opts.count
    if count:
        debug("adding %d elements to the queue...", count)
    else:
        debug("adding elements to the queue forever...")
    dq = new_dirq(1)
    if opts.type == "simple":
        element = ''
    else:
        element = {}
        if opts.header:
            element['header'] = dict(os.environ)
    done = 0
    time1 = time.time()
    while not count or done < count:
        done += 1
        if size:
            if opts.type == "simple":
                element = _body(size, rnd)
            else:
                element['body'] = _body(size, rnd)
        else:
            if opts.type == "simple":
                element = 'Element %i ;-)\n' % done
            else:
                try:
                    element['body'] = ('Élément %d \u263A\n' %
                                       done).decode("utf-8")
                except AttributeError:
                    element['body'] = 'Élément %d \u263A\n' % done
        _ = dq.add(element)
    time2 = time.time()
    debug("done in %.4f seconds", time2 - time1)


def test_remove():
    """Remove elements from the queue.
    """
    count = opts.count
    if count:
        debug("removing %d elements from the queue...", count)
    else:
        debug("removing all elements from the queue (one pass)...")
    dq = new_dirq(0)
    done = 0
    if count:
        # loop to iterate until enough are removed
        time1 = time.time()
        while done < count:
            name = dq.first()
            while name and done < count:
                if not dq.lock(name):
                    name = dq.next()
                    continue
                dq.remove(name)
                done += 1
                name = dq.next()
        time2 = time.time()
        debug("done in %.4f seconds", time2 - time1)
    else:
        # one pass only
        time1 = time.time()
        name = dq.first()
        while name:
            if not dq.lock(name):
                name = dq.next()
                continue
            dq.remove(name)
            done += 1
            name = dq.next()
        time2 = time.time()
        debug("done in %.4f seconds (%d elements removed)",
              time2 - time1, done)


def test_iterate():
    """Iterate through the queue (only lock+unlock).
    """
    dq = new_dirq(0)
    done = 0
    debug("iterating all elements in the queue (first()/next())...")
    time1 = time.time()
    name = dq.first()
    while name:
        if not dq.lock(name):
            name = dq.next()
            continue
        dq.unlock(name)
        done += 1
        name = dq.next()
    time2 = time.time()
    debug("done in %.4f seconds (%d elements)", time2 - time1, done)
    debug("iterating all elements in the queue (iterator protocol)...")
    time1 = time.time()
    for name in dq:
        if not dq.lock(name):
            continue
        dq.unlock(name)
    time2 = time.time()
    debug("done in %.4f seconds (%d elements)", time2 - time1, done)


def test_get():
    """Get all elements from the queue.
    """
    debug("getting all elements in the queue (one pass)...")
    dq = new_dirq(1)
    done = 0
    time1 = time.time()
    name = dq.first()
    while name:
        if not dq.lock(name):
            name = dq.next()
            continue
        data = dq.get(name)
        if opts.type == "simple" and len(data) < 10:
            _die("unexpected element %s: %s" % (name, data))
        dq.unlock(name)
        done += 1
        name = dq.next()
    time2 = time.time()
    debug("done in %.4f seconds (%d elements)", time2 - time1, done)


def test_simple():
    """Simple test filling and emptying a brand new queue.
    """
    path = opts.path
    if os.path.exists(path):
        _die("%s: directory exists: %s", ProgramName, opts.path)
    if not opts.count:
        _die("%s: missing option: -count", ProgramName)
    time1 = time.time()
    test_add()
    test_count()
    test_get()
    test_remove()
    test_purge()
    time2 = time.time()

    def directory_contents(path):
        """Wrapeed os.listdir()"""
        try:
            return os.listdir(path)
        except OSError:
            _die("%s: couldn't listdir(%s)", ProgramName, path)
    subdirs = directory_contents(path)
    if opts.type == "simple":
        num_subdirs = 1
    else:
        num_subdirs = 3
    if len(subdirs) != num_subdirs:
        _die("%s: unexpected subdirs: %i", ProgramName, len(subdirs))
    shutil.rmtree(path, ignore_errors=True)
    debug("done in %.4f seconds", time2 - time1)


if __name__ == "__main__":
    init()
    if TEST == 'all':
        TESTS.remove('all')
        tests = TESTS
    else:
        tests = [TEST]

    if opts.sleep:
        time.sleep(opts.sleep)
    for test in tests:
        test_func = 'test_%s()' % test
        exec(test_func)



# -*- coding: utf-8 -*-

import os
import shutil
import tempfile
import unittest

from dirq.queue import Queue
from dirq.Exceptions import QueueError
from dirq.QueueSet import QueueSet

__all__ = ['TestQueueSet']


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq-set')
        for i in range(1, 5):
            setattr(self, 'q%i' % i, '%s/%s' % (self.tempdir, 'q%i' % i))

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueueSet(TestDirQueue):
    """ TestQueueSet """

    def test1_init(self):
        """ QueueSet.__init__() """
        self.assertRaises(TypeError, QueueSet, ('a'))
        self.assertRaises(TypeError, QueueSet,
                          (Queue(self.q1, schema={'data': 'string'}), 'a'))
        self.assertRaises(TypeError, QueueSet,
                          ([Queue(self.q1, schema={'data': 'string'}), 'a']))
        self.assertRaises(TypeError, QueueSet, ([1, 2]))
        self.assertRaises(TypeError, QueueSet, ((1, 2)))

    def test2_addremove(self):
        """ QueueSet.add()/remove() """
        q1 = Queue(self.q1, schema={'data': 'string'})
        q2 = Queue(self.q2, schema={'data': 'string'})
        q3 = Queue(self.q3, schema={'data': 'string'})
        q4 = Queue(self.q4, schema={'data': 'string'})
        for i in range(10):
            q1.add({'data': '%i A\n' % i})
            q2.add({'data': '%i A\n' % i})
            q3.add({'data': '%i A\n' % i})
            q4.add({'data': '%i A\n' % i})
        qs = QueueSet([q1, q2])
        qs.add(q3)
        qs.remove(q1)
        qs.add(q1, q4)
        self.assertRaises(QueueError, qs.add, ([q1]))

    def test3_firstnext(self):
        """ QueueSet.first()/next() """
        q1 = Queue(self.q1, schema={'data': 'string'})
        q2 = Queue(self.q2, schema={'data': 'string'})
        for i in range(10):
            q1.add({'data': '%i A\n' % i})
            q2.add({'data': '%i A\n' % i})
        qs = QueueSet([q1, q2])
        e = qs.first()
        assert isinstance(e, tuple)
        assert isinstance(e[0], Queue)
        assert isinstance(e[1], str)
        while e[0]:
            e = qs.next()


def main():
    testcases = [TestQueueSet]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import shutil
import tempfile
import time
import unittest

from dirq.QueueSimple import QueueSimple, LOCKED_SUFFIX

__all__ = ['TestQueueSimple']


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq-simple')
        self.qdir = self.tempdir + '/dirq'

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueueSimple(TestDirQueue):

    def test01_init(self):
        'QueueSimple.__init__()'
        path = self.tempdir + '/aaa/bbb/ccc'
        granularity = 30
        qs = QueueSimple(path, granularity=granularity)
        assert qs.path == path
        assert qs.granularity == granularity

    def test02_add_data(self):
        'QueueSimple._add_data()'
        data = 'f0o'
        qs = QueueSimple(self.qdir)
        _subdirName, _fullFnOrig = qs._add_data(data)
        subdirs = os.listdir(self.qdir)
        assert len(subdirs) == 1
        assert _subdirName == subdirs[0]
        subdir = self.qdir + '/' + subdirs[0]
        files = os.listdir(subdir)
        assert len(files) == 1
        fn = subdir + '/' + files[0]
        assert _fullFnOrig == fn
        assert open(fn, 'r').read() == data

    def test03_add_path(self):
        'QueueSimple._add_path()'
        data = 'abc'
        qs = QueueSimple(self.qdir)
        _dir = 'elems'
        elems_dir = self.qdir + '/' + _dir
        os.mkdir(elems_dir)
        _tmpName = self.tempdir + '/elem.tmp'
        open(_tmpName, 'w').write(data)
        assert len(os.listdir(self.tempdir)) == 2
        newName = qs._add_path(_tmpName, _dir)
        assert len(os.listdir(elems_dir)) == 1
        assert len(os.listdir(self.tempdir)) == 1
        assert open(self.qdir + '/' + newName).read() == data

    def test04_add(self):
        'QueueSimple.add()'
        data = 'foo bar'
        qs = QueueSimple(self.qdir)
        elem = qs.add(data)
        assert open(self.qdir + '/' + elem).read() == data

    def test05_add_path(self):
        'QueueSimple.add_path()'
        qs = QueueSimple(self.qdir)
        data = 'foo0oo'
        path = self.tempdir + '/foo.bar'
        open(path, 'w').write(data)
        elem = qs.add_path(path)
        assert open(self.qdir + '/' + elem).read() == data
        self.assertFalse(os.path.exists(path))

    def test06_lock_unlock(self):
        'QueueSimple.lock()'
        qs = QueueSimple(self.qdir)
        data = 'foo'
        elem_name = 'foo.bar'
        elem_full_path = self.qdir + '/' + elem_name
        open(elem_full_path, 'w').write(data)
        self.assertEqual(qs.lock(elem_name), 1)
        self.assertTrue(os.path.exists(elem_full_path + LOCKED_SUFFIX))
        qs.unlock(elem_name)

    def test07_get(self):
        'QueueSimple.get()'
        data = 'foo'.encode()
        qs = QueueSimple(self.qdir)
        elem = qs.add(data)
        qs.lock(elem)
        self.assertEqual(qs.get(elem), data)

    def test08_count(self):
        'QueueSimple.count()'
        qs = QueueSimple(self.qdir)
        # add "normal" element
        qs.add('foo')
        # simply add a file (fake element) into the elements directory
        fake_elem = os.listdir(self.qdir)[0] + '/' + 'foo.bar'
        open(self.qdir + '/' + fake_elem, 'w').write('')
        self.assertEqual(qs.count(), 1)

    def test09_remove(self):
        'QueueSimple.remove()'
        qs = QueueSimple(self.qdir, granularity=1)
        for _ in range(5):
            qs.add('foo')
        assert qs.count() == 5
        for elem in qs:
            qs.lock(elem)
            qs.remove(elem)
        self.assertEqual(qs.count(), 0)

    def test10_purge_oneDirOneElement(self):
        'QueueSimple.purge() one directory & element'
        qs = QueueSimple(self.qdir)
        qs.add('foo')
        self.assertEqual(qs.count(), 1)
        elem = qs.first()
        qs.lock(elem)
        elem_path_lock = self.qdir + '/' + elem + LOCKED_SUFFIX
        self.assertTrue(os.path.exists(elem_path_lock) is True)
        time.sleep(2)
        qs.purge(maxlock=1)
        self.assertTrue(os.path.exists(elem_path_lock) is False)
        self.assertEqual(qs.count(), 1)
        self.assertEqual(len(os.listdir(self.qdir)), 1)

    def test11_purge_multDirMultElement(self):
        'QueueSimple.purge() multiple directories & elements'
        qs = QueueSimple(self.qdir, granularity=1)

        qs.add('foo')
        assert qs.count() == 1
        time.sleep(2)
        qs.add('bar')
        assert qs.count() == 2
        assert len(os.listdir(self.qdir)) == 2
        qs.purge()
        assert qs.count() == 2

        elem = qs.first()
        qs.lock(elem)
        qs.remove(elem)
        assert qs.count() == 1
        qs.purge()
        assert len(os.listdir(self.qdir)) == 1

        time.sleep(2)
        qs.add('baz')
        assert len(os.listdir(self.qdir)) == 2
        for elem in qs:
            qs.lock(elem)
        elem1 = qs.first()
        lock_path1 = self.qdir + '/' + elem1 + LOCKED_SUFFIX
        assert os.path.exists(lock_path1) is True
        os.utime(lock_path1, (time.time() - 25, time.time() - 25))
        qs.purge(maxlock=10)
        assert os.path.exists(lock_path1) is False

        elem2 = qs.next()
        lock_path2 = self.qdir + '/' + elem2 + LOCKED_SUFFIX
        assert os.path.exists(lock_path2) is True


def main():
    testcases = [TestQueueSimple]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



"""Exceptions used in the module.

Author
------

Konstantin Skaburskas <konstantin.skaburskas@gmail.com>

License and Copyright
---------------------

ASL 2.0

Copyright (C) CERN 2011-2021
"""


class QueueError(Exception):
    """QueueError"""


class QueueLockError(QueueError):
    """QueueLockError"""



# -*- coding: utf-8 -*-

import os
import shutil
import unittest

from dirq.QueueNull import QueueNull
import tempfile

__all__ = ['TestQueueNull']


def assert_fail(exc, callable, *args, **kwargs):
    try:
        callable(*args, **kwargs)
        raise AssertionError("An exception was expected")
    except exc:
        # fine
        pass


class TestDirQueue(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='dirq-null')
        self.qdir = self.tempdir + '/dirq'

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)


class TestQueueNull(TestDirQueue):

    def test01init(self):
        'QueueNull.__init__()'
        qn = QueueNull()

    def test02add(self):
        'QueueNull.add()'
        data = 'foo bar'
        qn = QueueNull()
        elem = qn.add(data)
        assert elem == ""

    def test03add_path(self):
        'QueueNull.add_path()'
        qn = QueueNull()
        data = 'foo'
        path = self.tempdir + '/foo.bar'
        fh = open(path, 'w')
        fh.write(data)
        fh.flush()
        fh.close()
        elem = qn.add_path(path)
        self.assertFalse(os.path.exists(path))

    def test04lockunlok(self):
        'QueueNull.lock()'
        qn = QueueNull()
        assert_fail(NotImplementedError, qn.lock, "")
        assert_fail(NotImplementedError, qn.unlock, "")

    def test05get(self):
        'QueueNull.get()'
        qn = QueueNull()
        assert_fail(NotImplementedError, qn.get, "")

    def test06count(self):
        'QueueNull.count()'
        qn = QueueNull()
        qn.add('foo1')
        assert qn.count() == 0
        qn.add('foo2')
        assert qn.count() == 0

    def test07remove(self):
        'QueueNull.remove()'
        qn = QueueNull()
        assert_fail(NotImplementedError, qn.remove, "")

    def test08purge_oneDirOneElement(self):
        'QueueNull.purge() one directory & element'
        qn = QueueNull()
        qn.purge()


def main():
    testcases = [TestQueueNull]
    for tc in testcases:
        unittest.TextTestRunner(verbosity=2).\
            run(unittest.TestLoader().loadTestsFromTestCase(tc))


if __name__ == "__main__":
    main()



import dirq

NAME = 'dirq'
VERSION = dirq.VERSION
DESCRIPTION = "Directory based queue"
LONG_DESCRIPTION = """
A port of Perl module Directory::Queue
http://search.cpan.org/dist/Directory-Queue/

The goal of this module is to offer a queue system using the underlying
filesystem for storage, security and to prevent race conditions via atomic
operations. It focuses on simplicity, robustness and scalability.

This module allows multiple concurrent readers and writers to interact with
the same queue."""
AUTHOR = 'Konstantin Skaburskas'
AUTHOR_EMAIL = 'konstantin.skaburskas@gmail.com'
LICENSE = "ASL 2.0"
PLATFORMS = "Any"
URL = "https://github.com/cern-mig/python-dirq"
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: Unix",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.4",
    "Programming Language :: Python :: 2.5",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.0",
    "Programming Language :: Python :: 3.1",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Topic :: Software Development :: Libraries :: Python Modules"
]

from distutils.core import setup, Command


class test(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from test import run_tests
        run_tests.main()

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      platforms=PLATFORMS,
      url=URL,
      classifiers=CLASSIFIERS,
      packages=['dirq'],
      cmdclass={'test': test}, )
